@@ -4,6 +4,7 @@
 """Wrapper c_api of LightGBM"""
 from __future__ import absolute_import
 
+import copy
 import ctypes
 import os
 import warnings
@@ -66,7 +67,7 @@ def is_numpy_1d_array(data):
 def is_1d_list(data):
     """Check is 1d list"""
     return isinstance(data, list) and \
-        (not data or isinstance(data[0], numeric_types))
+        (not data or is_numeric(data[0]))
 
 
 def list_to_1d_numpy(data, dtype=np.float32, name='list'):
@@ -224,10 +225,6 @@ PANDAS_DTYPE_MAPPER = {'int8': 'int', 'int16': 'int', 'int32': 'int',
 def _data_from_pandas(data, feature_name, categorical_feature, pandas_categorical):
     if isinstance(data, DataFrame):
         if feature_name == 'auto' or feature_name is None:
-            if all([isinstance(name, integer_types + (np.integer, )) for name in data.columns]):
-                msg = """Using Pandas (default) integer column names, not column indexes. You can use indexes with DataFrame.values."""
-                warnings.filterwarnings('once')
-                warnings.warn(msg, stacklevel=5)
             data = data.rename(columns=str)
         cat_cols = data.select_dtypes(include=['category']).columns
         if pandas_categorical is None:  # train dataset
@@ -595,11 +592,12 @@ class Dataset(object):
         self.silent = silent
         self.feature_name = feature_name
         self.categorical_feature = categorical_feature
-        self.params = params
+        self.params = copy.deepcopy(params)
         self.free_raw_data = free_raw_data
         self.used_indices = None
         self._predictor = None
         self.pandas_categorical = None
+        self.params_back_up = None
 
     def __del__(self):
         self._free_handle()
@@ -624,6 +622,8 @@ class Dataset(object):
         self.max_bin = max_bin
         self.predictor = predictor
         params["max_bin"] = max_bin
+        if "verbosity" in params:
+            params.setdefault("verbose", params.pop("verbosity"))
         if silent:
             params["verbose"] = 0
         elif "verbose" not in params:
@@ -642,8 +642,8 @@ class Dataset(object):
                 else:
                     raise TypeError("Wrong type({}) or unknown name({}) in categorical_feature"
                                     .format(type(name).__name__, name))
-
-            params['categorical_column'] = sorted(categorical_indices)
+            if categorical_indices:
+                params['categorical_column'] = sorted(categorical_indices)
 
         params_str = param_dict_to_str(params)
         """process for reference dataset"""
@@ -848,6 +848,8 @@ class Dataset(object):
         params : dict
             Other parameters
         """
+        if params is None:
+            params = self.params
         ret = Dataset(None, reference=self, feature_name=self.feature_name,
                       categorical_feature=self.categorical_feature, params=params)
         ret._predictor = self._predictor
@@ -872,8 +874,13 @@ class Dataset(object):
         if not self.params:
             self.params = params
         else:
+            self.params_back_up = copy.deepcopy(self.params)
             self.params.update(params)
 
+    def _reverse_update_params(self):
+        self.params = copy.deepcopy(self.params_back_up)
+        self.params_back_up = None
+
     def set_field(self, field_name, data):
         """Set property into the Dataset.
 
@@ -972,8 +979,15 @@ class Dataset(object):
         if self.categorical_feature == categorical_feature:
             return
         if self.data is not None:
-            self.categorical_feature = categorical_feature
-            self._free_handle()
+            if self.categorical_feature is None:
+                self.categorical_feature = categorical_feature
+                self._free_handle()
+            elif categorical_feature == 'auto':
+                warnings.warn('Using categorical_feature in Dataset.')
+            else:
+                warnings.warn('categorical_feature in Dataset is overrided. New categorical_feature is {}'.format(sorted(list(categorical_feature))))
+                self.categorical_feature = categorical_feature
+                self._free_handle()
         else:
             raise LightGBMError("Cannot set categorical feature after freed raw data, set free_raw_data=False when construct Dataset to avoid this.")
 
@@ -1002,7 +1016,8 @@ class Dataset(object):
         self.set_categorical_feature(reference.categorical_feature)
         self.set_feature_name(reference.feature_name)
         self._set_predictor(reference._predictor)
-        if self.reference is reference:
+        # we're done if self and reference share a common upstrem reference
+        if self.get_ref_chain().intersection(reference.get_ref_chain()):
             return
         if self.data is not None:
             self.reference = reference
@@ -1172,6 +1187,28 @@ class Dataset(object):
         else:
             raise LightGBMError("Cannot get num_feature before construct dataset")
 
+    def get_ref_chain(self, ref_limit=100):
+        '''
+        Gets a chain of Dataset objects, starting with r, then going to r.reference if exists,
+            then to r.reference.reference, etc. until we hit ref_limit or a reference loop
+
+        Returns
+        -------
+        chain of references of self : set of Dataset objects
+        '''
+        head = self
+        ref_chain = set()
+        while len(ref_chain) < ref_limit:
+            if isinstance(head, Dataset):
+                ref_chain.add(head)
+                if (head.reference is not None) and (head.reference not in ref_chain):
+                    head = head.reference
+                else:
+                    break
+            else:
+                break
+        return(ref_chain)
+
 
 class Booster(object):
     """"Booster in LightGBM."""
@@ -1190,13 +1227,15 @@ class Booster(object):
         silent : boolean, optional
             Whether print messages during construction
         """
-        self.handle = ctypes.c_void_p()
+        self.handle = None
         self.__need_reload_eval_info = True
         self.__train_data_name = "training"
         self.__attr = {}
         self.best_iteration = -1
         self.best_score = {}
         params = {} if params is None else params
+        if "verbosity" in params:
+            params.setdefault("verbose", params.pop("verbosity"))
         if silent:
             params["verbose"] = 0
         elif "verbose" not in params:
@@ -1207,6 +1246,7 @@ class Booster(object):
                 raise TypeError('Training data should be Dataset instance, met {}'.format(type(train_set).__name__))
             params_str = param_dict_to_str(params)
             """construct booster object"""
+            self.handle = ctypes.c_void_p()
             _safe_call(_LIB.LGBM_BoosterCreate(
                 train_set.construct().handle,
                 c_str(params_str),
@@ -1234,6 +1274,7 @@ class Booster(object):
         elif model_file is not None:
             """Prediction task"""
             out_num_iterations = ctypes.c_int(0)
+            self.handle = ctypes.c_void_p()
             _safe_call(_LIB.LGBM_BoosterCreateFromModelfile(
                 c_str(model_file),
                 ctypes.byref(out_num_iterations),
@@ -1245,7 +1286,7 @@ class Booster(object):
             self.__num_class = out_num_class.value
             self.pandas_categorical = _load_pandas_categorical(model_file)
         elif 'model_str' in params:
-            self.__load_model_from_string(params['model_str'])
+            self._load_model_from_string(params['model_str'])
         else:
             raise TypeError('Need at least one training dataset or model file to create booster instance')
 
@@ -1257,7 +1298,7 @@ class Booster(object):
         return self.__deepcopy__(None)
 
     def __deepcopy__(self, _):
-        model_str = self.__save_model_to_string()
+        model_str = self._save_model_to_string()
         booster = Booster({'model_str': model_str})
         booster.pandas_categorical = self.pandas_categorical
         return booster
@@ -1268,7 +1309,7 @@ class Booster(object):
         this.pop('train_set', None)
         this.pop('valid_sets', None)
         if handle is not None:
-            this["handle"] = self.__save_model_to_string()
+            this["handle"] = self._save_model_to_string()
         return this
 
     def __setstate__(self, state):
@@ -1286,6 +1327,11 @@ class Booster(object):
     def free_dataset(self):
         self.__dict__.pop('train_set', None)
         self.__dict__.pop('valid_sets', None)
+        self.__num_dataset = 0
+
+    def _free_buffer(self):
+        self.__inner_predict_buffer = []
+        self.__is_predicted_cur_iter = []
 
     def set_train_data_name(self, name):
         self.__train_data_name = name
@@ -1505,8 +1551,12 @@ class Booster(object):
             c_str(filename)))
         _save_pandas_categorical(filename, self.pandas_categorical)
 
-    def __load_model_from_string(self, model_str):
+    def _load_model_from_string(self, model_str):
         """[Private] Load model from string"""
+        if self.handle is not None:
+            _safe_call(_LIB.LGBM_BoosterFree(self.handle))
+        self._free_buffer()
+        self.handle = ctypes.c_void_p()
         out_num_iterations = ctypes.c_int(0)
         _safe_call(_LIB.LGBM_BoosterLoadModelFromString(
             c_str(model_str),
@@ -1518,7 +1568,7 @@ class Booster(object):
             ctypes.byref(out_num_class)))
         self.__num_class = out_num_class.value
 
-    def __save_model_to_string(self, num_iteration=-1):
+    def _save_model_to_string(self, num_iteration=-1):
         """[Private] Save model to string"""
         if num_iteration <= 0:
             num_iteration = self.best_iteration
@@ -1615,6 +1665,15 @@ class Booster(object):
             num_iteration = self.best_iteration
         return predictor.predict(data, num_iteration, raw_score, pred_leaf, data_has_header, is_reshape)
 
+    def get_leaf_output(self, tree_id, leaf_id):
+        ret = ctypes.c_double(0)
+        _safe_call(_LIB.LGBM_BoosterGetLeafValue(
+            self.handle,
+            ctypes.c_int(tree_id),
+            ctypes.c_int(leaf_id),
+            ctypes.byref(ret)))
+        return ret.value
+
     def _to_predictor(self, pred_parameter=None):
         """Convert to predictor"""
         predictor = _InnerPredictor(booster_handle=self.handle, pred_parameter=pred_parameter)
@@ -1769,7 +1828,7 @@ class Booster(object):
                 self.__name_inner_eval = \
                     [string_buffers[i].value.decode() for i in range_(self.__num_inner_eval)]
                 self.__higher_better_inner_eval = \
-                    [name.startswith(('auc', 'ndcg')) for name in self.__name_inner_eval]
+                    [name.startswith(('auc', 'ndcg', 'map')) for name in self.__name_inner_eval]
 
     def attr(self, key):
         """
