@@ -140,11 +140,11 @@ std::vector<double> GreedyFindBin(const double* distinct_values, const int* coun
   return bin_upper_bound;
 }
 
-std::vector<double> FindBinWithZeroAsMissing(const double* distinct_values, const int* counts,
+std::vector<double> FindBinWithZeroAsOneBin(const double* distinct_values, const int* counts,
                                              int num_distinct_values, int max_bin, size_t total_sample_cnt, int min_data_in_bin) {
   std::vector<double> bin_upper_bound;
   int left_cnt_data = 0;
-  int cnt_missing = 0;
+  int cnt_zero = 0;
   int right_cnt_data = 0;
   for (int i = 0; i < num_distinct_values; ++i) {
     if (distinct_values[i] <= -kZeroAsMissingValueRange) {
@@ -152,11 +152,11 @@ std::vector<double> FindBinWithZeroAsMissing(const double* distinct_values, cons
     } else if (distinct_values[i] > kZeroAsMissingValueRange) {
       right_cnt_data += counts[i];
     } else {
-      cnt_missing += counts[i];
+      cnt_zero += counts[i];
     }
   }
 
-  int left_cnt = 0;
+  int left_cnt = -1;
   for (int i = 0; i < num_distinct_values; ++i) {
     if (distinct_values[i] > -kZeroAsMissingValueRange) {
       left_cnt = i;
@@ -164,8 +164,12 @@ std::vector<double> FindBinWithZeroAsMissing(const double* distinct_values, cons
     }
   }
 
+  if (left_cnt < 0) {
+    left_cnt = num_distinct_values;
+  }
+
   if (left_cnt > 0) {
-    int left_max_bin = static_cast<int>(static_cast<double>(left_cnt_data) / (total_sample_cnt - cnt_missing) * (max_bin - 1));
+    int left_max_bin = static_cast<int>(static_cast<double>(left_cnt_data) / (total_sample_cnt - cnt_zero) * (max_bin - 1));
     bin_upper_bound = GreedyFindBin(distinct_values, counts, left_cnt, left_max_bin, left_cnt_data, min_data_in_bin);
     bin_upper_bound.back() = -kZeroAsMissingValueRange;
   }
@@ -255,19 +259,16 @@ void BinMapper::FindBin(double* values, int num_sample_values, size_t total_samp
   max_val_ = distinct_values.back();
   std::vector<int> cnt_in_bin;
   int num_distinct_values = static_cast<int>(distinct_values.size());
-  if (num_distinct_values + (na_cnt > 0) <= 2) {
-    bin_type_ = BinType::NumericalBin;
-  }
   if (bin_type_ == BinType::NumericalBin) {
     if (missing_type_ == MissingType::Zero) {
-      bin_upper_bound_ = FindBinWithZeroAsMissing(distinct_values.data(), counts.data(), num_distinct_values, max_bin, total_sample_cnt, min_data_in_bin);
+      bin_upper_bound_ = FindBinWithZeroAsOneBin(distinct_values.data(), counts.data(), num_distinct_values, max_bin, total_sample_cnt, min_data_in_bin);
       if (bin_upper_bound_.size() == 2) {
         missing_type_ = MissingType::None;
       }
     } else if (missing_type_ == MissingType::None) {
-      bin_upper_bound_ = GreedyFindBin(distinct_values.data(), counts.data(), num_distinct_values, max_bin, total_sample_cnt, min_data_in_bin);
+      bin_upper_bound_ = FindBinWithZeroAsOneBin(distinct_values.data(), counts.data(), num_distinct_values, max_bin, total_sample_cnt, min_data_in_bin);
     } else {
-      bin_upper_bound_ = GreedyFindBin(distinct_values.data(), counts.data(), num_distinct_values, max_bin - 1, total_sample_cnt - na_cnt, min_data_in_bin);
+      bin_upper_bound_ = FindBinWithZeroAsOneBin(distinct_values.data(), counts.data(), num_distinct_values, max_bin - 1, total_sample_cnt - na_cnt, min_data_in_bin);
       bin_upper_bound_.push_back(NaN);
     }
     num_bin_ = static_cast<int>(bin_upper_bound_.size());
@@ -307,32 +308,46 @@ void BinMapper::FindBin(double* values, int num_sample_values, size_t total_samp
       std::swap(distinct_values_int[0], distinct_values_int[1]);
     }
     // will ignore the categorical of small counts
-    const int cut_cnt = static_cast<int>((total_sample_cnt - na_cnt) * 0.99f);
+    int cut_cnt = static_cast<int>((total_sample_cnt - na_cnt) * 0.99f);
+    size_t cur_cat = 0;
     categorical_2_bin_.clear();
     bin_2_categorical_.clear();
     num_bin_ = 0;
     int used_cnt = 0;
     max_bin = std::min(static_cast<int>(distinct_values_int.size()), max_bin);
-    while (used_cnt < cut_cnt || num_bin_ < max_bin) {
-      if (distinct_values_int[num_bin_] < 0) {
-        Log::Fatal("Cannot use negative numbers in categorcial features.");
+    cnt_in_bin.clear();
+    while (cur_cat < distinct_values_int.size() 
+           && (used_cnt < cut_cnt || num_bin_ < max_bin)) {
+      if (distinct_values_int[cur_cat] < 0) {
+        na_cnt += counts_int[cur_cat];
+        cut_cnt -= counts_int[cur_cat];
+        Log::Warning("Met negative value in categorical features, will convert it to NaN");
+      } else {
+        bin_2_categorical_.push_back(distinct_values_int[cur_cat]);
+        categorical_2_bin_[distinct_values_int[cur_cat]] = static_cast<unsigned int>(num_bin_);
+        used_cnt += counts_int[cur_cat];
+        cnt_in_bin.push_back(counts_int[cur_cat]);
+        ++num_bin_;
       }
-      bin_2_categorical_.push_back(distinct_values_int[num_bin_]);
-      categorical_2_bin_[distinct_values_int[num_bin_]] = static_cast<unsigned int>(num_bin_);
-      used_cnt += counts_int[num_bin_];
+      ++cur_cat;
+    }
+    // need an additional bin for NaN
+    if (cur_cat == distinct_values_int.size() && na_cnt > 0) {
+      // use -1 to represent NaN
+      bin_2_categorical_.push_back(-1);
+      categorical_2_bin_[-1] = num_bin_;
+      cnt_in_bin.push_back(0);
       ++num_bin_;
     }
     // Use MissingType::None to represent this bin contains all categoricals
-    if (num_bin_ == static_cast<int>(distinct_values_int.size()) && na_cnt == 0) {
+    if (cur_cat == distinct_values_int.size() && na_cnt == 0) {
       missing_type_ = MissingType::None;
     } else if (na_cnt == 0) {
       missing_type_ = MissingType::Zero;
     } else {
       missing_type_ = MissingType::NaN;
     }
-    cnt_in_bin = counts_int;
-    counts_int.resize(num_bin_);
-    counts_int.back() += static_cast<int>(total_sample_cnt - used_cnt);
+    cnt_in_bin.back() += static_cast<int>(total_sample_cnt - used_cnt);
   }
 
   // check trival(num_bin_ == 1) feature
