@@ -32,49 +32,10 @@ std::unordered_map<std::string, std::string> ConfigBase::Str2Map(const char* par
   return params;
 }
 
-void OverallConfig::Set(const std::unordered_map<std::string, std::string>& params) {
-  // load main config types
-  GetInt(params, "num_threads", &num_threads);
-  GetString(params, "convert_model_language", &convert_model_language);
-
-  // generate seeds by seed.
-  if (GetInt(params, "seed", &seed)) {
-    Random rand(seed);
-    int int_max = std::numeric_limits<short>::max();
-    io_config.data_random_seed = static_cast<int>(rand.NextShort(0, int_max));
-    boosting_config.bagging_seed = static_cast<int>(rand.NextShort(0, int_max));
-    boosting_config.drop_seed = static_cast<int>(rand.NextShort(0, int_max));
-    boosting_config.tree_config.feature_fraction_seed = static_cast<int>(rand.NextShort(0, int_max));
-  }
-  GetTaskType(params);
-  GetBoostingType(params);
-  GetObjectiveType(params);
-  GetMetricType(params);
-
-  // sub-config setup
-  network_config.Set(params);
-  io_config.Set(params);
-
-  boosting_config.Set(params);
-  objective_config.Set(params);
-  metric_config.Set(params);
-  // check for conflicts
-  CheckParamConflict();
-
-  if (io_config.verbosity == 1) {
-    LightGBM::Log::ResetLogLevel(LightGBM::LogLevel::Info);
-  } else if (io_config.verbosity == 0) {
-    LightGBM::Log::ResetLogLevel(LightGBM::LogLevel::Warning);
-  } else if (io_config.verbosity >= 2) {
-    LightGBM::Log::ResetLogLevel(LightGBM::LogLevel::Debug);
-  } else {
-    LightGBM::Log::ResetLogLevel(LightGBM::LogLevel::Fatal);
-  }
-}
-
-void OverallConfig::GetBoostingType(const std::unordered_map<std::string, std::string>& params) {
+std::string GetBoostingType(const std::unordered_map<std::string, std::string>& params) {
   std::string value;
-  if (GetString(params, "boosting_type", &value)) {
+  std::string boosting_type = kDefaultBoostingType;
+  if (ConfigBase::GetString(params, "boosting_type", &value)) {
     std::transform(value.begin(), value.end(), value.begin(), Common::tolower);
     if (value == std::string("gbdt") || value == std::string("gbrt")) {
       boosting_type = "gbdt";
@@ -82,23 +43,29 @@ void OverallConfig::GetBoostingType(const std::unordered_map<std::string, std::s
       boosting_type = "dart";
     } else if (value == std::string("goss")) {
       boosting_type = "goss";
+    } else if (value == std::string("rf") || value == std::string("randomforest")) {
+      boosting_type = "rf";
     } else {
       Log::Fatal("Unknown boosting type %s", value.c_str());
     }
   }
+  return boosting_type;
 }
 
-void OverallConfig::GetObjectiveType(const std::unordered_map<std::string, std::string>& params) {
+std::string GetObjectiveType(const std::unordered_map<std::string, std::string>& params) {
   std::string value;
-  if (GetString(params, "objective", &value)) {
+  std::string objective_type = kDefaultObjectiveType;
+  if (ConfigBase::GetString(params, "objective", &value)) {
     std::transform(value.begin(), value.end(), value.begin(), Common::tolower);
     objective_type = value;
   }
+  return objective_type;
 }
 
-void OverallConfig::GetMetricType(const std::unordered_map<std::string, std::string>& params) {
+std::vector<std::string> GetMetricType(const std::unordered_map<std::string, std::string>& params) {
   std::string value;
-  if (GetString(params, "metric", &value)) {
+  std::vector<std::string> metric_types;
+  if (ConfigBase::GetString(params, "metric", &value)) {
     // clear old metrics
     metric_types.clear();
     // to lower
@@ -118,17 +85,18 @@ void OverallConfig::GetMetricType(const std::unordered_map<std::string, std::str
     }
     metric_types.shrink_to_fit();
   }
+  return metric_types;
 }
 
-
-void OverallConfig::GetTaskType(const std::unordered_map<std::string, std::string>& params) {
+TaskType GetTaskType(const std::unordered_map<std::string, std::string>& params) {
   std::string value;
-  if (GetString(params, "task", &value)) {
+  TaskType task_type = TaskType::kTrain;
+  if (ConfigBase::GetString(params, "task", &value)) {
     std::transform(value.begin(), value.end(), value.begin(), Common::tolower);
     if (value == std::string("train") || value == std::string("training")) {
       task_type = TaskType::kTrain;
     } else if (value == std::string("predict") || value == std::string("prediction")
-      || value == std::string("test")) {
+               || value == std::string("test")) {
       task_type = TaskType::kPredict;
     } else if (value == std::string("convert_model")) {
       task_type = TaskType::kConvertModel;
@@ -136,12 +104,90 @@ void OverallConfig::GetTaskType(const std::unordered_map<std::string, std::strin
       Log::Fatal("Unknown task type %s", value.c_str());
     }
   }
+  return task_type;
+}
+
+std::string GetDeviceType(const std::unordered_map<std::string, std::string>& params) {
+  std::string value;
+  std::string device_type = kDefaultDevice;
+  if (ConfigBase::GetString(params, "device", &value)) {
+    std::transform(value.begin(), value.end(), value.begin(), Common::tolower);
+    if (value == std::string("cpu")) {
+      device_type = "cpu";
+    } else if (value == std::string("gpu")) {
+      device_type = "gpu";
+    } else {
+      Log::Fatal("Unknown device type %s", value.c_str());
+    }
+  }
+  return device_type;
+}
+
+std::string GetTreeLearnerType(const std::unordered_map<std::string, std::string>& params) {
+  std::string value;
+  std::string tree_learner_type = kDefaultTreeLearnerType;
+  if (ConfigBase::GetString(params, "tree_learner", &value)) {
+    std::transform(value.begin(), value.end(), value.begin(), Common::tolower);
+    if (value == std::string("serial")) {
+      tree_learner_type = "serial";
+    } else if (value == std::string("feature") || value == std::string("feature_parallel")) {
+      tree_learner_type = "feature";
+    } else if (value == std::string("data") || value == std::string("data_parallel")) {
+      tree_learner_type = "data";
+    } else if (value == std::string("voting") || value == std::string("voting_parallel")) {
+      tree_learner_type = "voting";
+    } else {
+      Log::Fatal("Unknown tree learner type %s", value.c_str());
+    }
+  }
+  return tree_learner_type;
+}
+
+void OverallConfig::Set(const std::unordered_map<std::string, std::string>& params) {
+  // load main config types
+  GetInt(params, "num_threads", &num_threads);
+  GetString(params, "convert_model_language", &convert_model_language);
+
+  // generate seeds by seed.
+  if (GetInt(params, "seed", &seed)) {
+    Random rand(seed);
+    int int_max = std::numeric_limits<short>::max();
+    io_config.data_random_seed = static_cast<int>(rand.NextShort(0, int_max));
+    boosting_config.bagging_seed = static_cast<int>(rand.NextShort(0, int_max));
+    boosting_config.drop_seed = static_cast<int>(rand.NextShort(0, int_max));
+    boosting_config.tree_config.feature_fraction_seed = static_cast<int>(rand.NextShort(0, int_max));
+  }
+  task_type = GetTaskType(params);
+  boosting_type = GetBoostingType(params);
+
+  metric_types = GetMetricType(params);
+
+  // sub-config setup
+  network_config.Set(params);
+  io_config.Set(params);
+
+  boosting_config.Set(params);
+  objective_type = GetObjectiveType(params);
+  objective_config.Set(params);
+  metric_config.Set(params);
+
+  // check for conflicts
+  CheckParamConflict();
+
+  if (io_config.verbosity == 1) {
+    LightGBM::Log::ResetLogLevel(LightGBM::LogLevel::Info);
+  } else if (io_config.verbosity == 0) {
+    LightGBM::Log::ResetLogLevel(LightGBM::LogLevel::Warning);
+  } else if (io_config.verbosity >= 2) {
+    LightGBM::Log::ResetLogLevel(LightGBM::LogLevel::Debug);
+  } else {
+    LightGBM::Log::ResetLogLevel(LightGBM::LogLevel::Fatal);
+  }
 }
 
 void OverallConfig::CheckParamConflict() {
-
   // check if objective_type, metric_type, and num_class match
-  bool objective_type_multiclass = (objective_type == std::string("multiclass") 
+  bool objective_type_multiclass = (objective_type == std::string("multiclass")
                                     || objective_type == std::string("multiclassova"));
   int num_class_check = boosting_config.num_class;
   if (objective_type_multiclass) {
@@ -155,7 +201,7 @@ void OverallConfig::CheckParamConflict() {
   }
   if (boosting_config.is_provide_training_metric || !io_config.valid_data_filenames.empty()) {
     for (std::string metric_type : metric_types) {
-      bool metric_type_multiclass = (metric_type == std::string("multi_logloss") 
+      bool metric_type_multiclass = (metric_type == std::string("multi_logloss")
                                      || metric_type == std::string("multi_error"));
       if ((objective_type_multiclass && !metric_type_multiclass)
         || (!objective_type_multiclass && metric_type_multiclass)) {
@@ -171,25 +217,33 @@ void OverallConfig::CheckParamConflict() {
     boosting_config.tree_learner_type = "serial";
   }
 
-  if (boosting_config.tree_learner_type == std::string("serial")) {
+  bool is_single_tree_learner = boosting_config.tree_learner_type == std::string("serial");
+
+  if (is_single_tree_learner) {
     is_parallel = false;
     network_config.num_machines = 1;
   }
 
-  if (boosting_config.tree_learner_type == std::string("serial") 
-      || boosting_config.tree_learner_type == std::string("feature")) {
+  if (is_single_tree_learner || boosting_config.tree_learner_type == std::string("feature")) {
     is_parallel_find_bin = false;
   } else if (boosting_config.tree_learner_type == std::string("data")
              || boosting_config.tree_learner_type == std::string("voting")) {
     is_parallel_find_bin = true;
-    if (boosting_config.tree_config.histogram_pool_size >= 0 
+    if (boosting_config.tree_config.histogram_pool_size >= 0
         && boosting_config.tree_learner_type == std::string("data")) {
       Log::Warning("Histogram LRU queue was enabled (histogram_pool_size=%f). Will disable this to reduce communication costs"
         , boosting_config.tree_config.histogram_pool_size);
       // Change pool size to -1 (no limit) when using data parallel to reduce communication costs
       boosting_config.tree_config.histogram_pool_size = -1;
     }
-
+  }
+  // Check max_depth and num_leaves
+  if (boosting_config.tree_config.max_depth > 0) {
+    int full_num_leaves = static_cast<int>(std::pow(2, boosting_config.tree_config.max_depth));
+    if (full_num_leaves > boosting_config.tree_config.num_leaves 
+        && boosting_config.tree_config.num_leaves == kDefaultNumLeaves) {
+      Log::Warning("Accuarcy may be bad since you didn't set num_leaves.");
+    }
   }
 }
 
@@ -199,6 +253,7 @@ void IOConfig::Set(const std::unordered_map<std::string, std::string>& params) {
   GetInt(params, "num_class", &num_class);
   GetInt(params, "data_random_seed", &data_random_seed);
   GetString(params, "data", &data_filename);
+  GetString(params, "init_score_file", &initscore_filename);
   GetInt(params, "verbose", &verbosity);
   GetInt(params, "num_iteration_predict", &num_iteration_predict);
   GetInt(params, "bin_construct_sample_cnt", &bin_construct_sample_cnt);
@@ -210,6 +265,7 @@ void IOConfig::Set(const std::unordered_map<std::string, std::string>& params) {
   GetBool(params, "enable_load_from_binary_file", &enable_load_from_binary_file);
   GetBool(params, "is_predict_raw_score", &is_predict_raw_score);
   GetBool(params, "is_predict_leaf_index", &is_predict_leaf_index);
+  GetBool(params, "is_predict_contrib", &is_predict_contrib);
   GetInt(params, "snapshot_freq", &snapshot_freq);
   GetString(params, "output_model", &output_model);
   GetString(params, "input_model", &input_model);
@@ -219,6 +275,12 @@ void IOConfig::Set(const std::unordered_map<std::string, std::string>& params) {
   if (GetString(params, "valid_data", &tmp_str)) {
     valid_data_filenames = Common::Split(tmp_str.c_str(), ',');
   }
+  if (GetString(params, "valid_init_score_file", &tmp_str)) {
+    valid_data_initscores = Common::Split(tmp_str.c_str(), ',');
+  } else {
+    valid_data_initscores = std::vector<std::string>(valid_data_filenames.size(), "");
+  }
+  CHECK(valid_data_filenames.size() == valid_data_initscores.size());
   GetBool(params, "has_header", &has_header);
   GetString(params, "label_column", &label_column);
   GetString(params, "weight_column", &weight_column);
@@ -234,22 +296,9 @@ void IOConfig::Set(const std::unordered_map<std::string, std::string>& params) {
   GetBool(params, "pred_early_stop", &pred_early_stop);
   GetInt(params, "pred_early_stop_freq", &pred_early_stop_freq);
   GetDouble(params, "pred_early_stop_margin", &pred_early_stop_margin);
-
-  GetDeviceType(params);
-}
-
-void IOConfig::GetDeviceType(const std::unordered_map<std::string, std::string>& params) {
-  std::string value;
-  if (GetString(params, "device", &value)) {
-    std::transform(value.begin(), value.end(), value.begin(), Common::tolower);
-    if (value == std::string("cpu")) {
-      device_type = "cpu";
-    } else if (value == std::string("gpu")) {
-      device_type = "gpu";
-    } else {
-      Log::Fatal("Unknown device type %s", value.c_str());
-    }
-  }
+  GetBool(params, "use_missing", &use_missing);
+  GetBool(params, "zero_as_missing", &zero_as_missing);
+  device_type = GetDeviceType(params);
 }
 
 void ObjectiveConfig::Set(const std::unordered_map<std::string, std::string>& params) {
@@ -333,10 +382,18 @@ void TreeConfig::Set(const std::unordered_map<std::string, std::string>& params)
   GetInt(params, "gpu_platform_id", &gpu_platform_id);
   GetInt(params, "gpu_device_id", &gpu_device_id);
   GetBool(params, "gpu_use_dp", &gpu_use_dp);
-  GetBool(params, "use_missing", &use_missing);
+  GetInt(params, "max_cat_group", &max_cat_group);
+  GetInt(params, "max_cat_threshold", &max_cat_threshold);
+  GetDouble(params, "cat_smooth_ratio", &cat_smooth_ratio);
+  GetDouble(params, "min_cat_smooth", &min_cat_smooth);
+  GetDouble(params, "max_cat_smooth", &max_cat_smooth);
+  CHECK(max_cat_group > 1);
+  CHECK(max_cat_threshold > 0);
+  CHECK(cat_smooth_ratio >= 0);
+  CHECK(min_cat_smooth >= 0);
+  CHECK(max_cat_smooth > min_cat_smooth);
 }
 
-
 void BoostingConfig::Set(const std::unordered_map<std::string, std::string>& params) {
   GetInt(params, "num_iterations", &num_iterations);
   GetDouble(params, "sigmoid", &sigmoid);
@@ -365,43 +422,11 @@ void BoostingConfig::Set(const std::unordered_map<std::string, std::string>& par
   GetBool(params, "boost_from_average", &boost_from_average);
   CHECK(drop_rate <= 1.0 && drop_rate >= 0.0);
   CHECK(skip_drop <= 1.0 && skip_drop >= 0.0);
-  GetDeviceType(params);
-  GetTreeLearnerType(params);
+  device_type = GetDeviceType(params);
+  tree_learner_type = GetTreeLearnerType(params);
   tree_config.Set(params);
 }
 
-void BoostingConfig::GetTreeLearnerType(const std::unordered_map<std::string, std::string>& params) {
-  std::string value;
-  if (GetString(params, "tree_learner", &value)) {
-    std::transform(value.begin(), value.end(), value.begin(), Common::tolower);
-    if (value == std::string("serial")) {
-      tree_learner_type = "serial";
-    } else if (value == std::string("feature") || value == std::string("feature_parallel")) {
-      tree_learner_type = "feature";
-    } else if (value == std::string("data") || value == std::string("data_parallel")) {
-      tree_learner_type = "data";
-    } else if (value == std::string("voting") || value == std::string("voting_parallel")) {
-      tree_learner_type = "voting";
-    } else {
-      Log::Fatal("Unknown tree learner type %s", value.c_str());
-    }
-  }
-}
-
-void BoostingConfig::GetDeviceType(const std::unordered_map<std::string, std::string>& params) {
-  std::string value;
-  if (GetString(params, "device", &value)) {
-    std::transform(value.begin(), value.end(), value.begin(), Common::tolower);
-    if (value == std::string("cpu")) {
-      device_type = "cpu";
-    } else if (value == std::string("gpu")) {
-      device_type = "gpu";
-    } else {
-      Log::Fatal("Unknown device type %s", value.c_str());
-    }
-  }
-}
-
 void NetworkConfig::Set(const std::unordered_map<std::string, std::string>& params) {
   GetInt(params, "num_machines", &num_machines);
   CHECK(num_machines >= 1);
